@@ -10,21 +10,29 @@ from .preprocess import Preprocessor
 from .parser_ import Parser
 from .optimizer import Optimizer
 from .linker import Linker
-from .error import MlogError
+from .error import Error
+from .compile import compile_code
 from . import __version__
 
 
-# input/output method
 class IOMethod(enum.Enum):
+    """
+    Method for code input/output.
+    """
+
     FILE = enum.auto()
     STD = enum.auto()
     CLIP = enum.auto()
 
 
-def main():
+def main() -> None:
+    """
+    Parse command line arguments and compile code.
+    """
+
     parser = argparse.ArgumentParser(description="Mindustry logic compiler", prog="mlog++")
 
-    parser.add_argument("file", type=str, help="input file(s) [@clip for clipboard]", nargs="+")
+    parser.add_argument("file", type=str, help="input file [@clip for clipboard]")
 
     parser.add_argument("-o:f", "--output-file", help="write output to a file")
     parser.add_argument("-o:s", "--output-stdout", help="write output to stdout", action="store_true")
@@ -39,6 +47,7 @@ def main():
 
     args = parser.parse_args()
 
+    # default output method
     output_method = IOMethod.CLIP
     output_file = ""
 
@@ -59,53 +68,28 @@ def main():
 
                 verbose = v
 
-    # check if files exist
-    for fn in args.file:
-        # exists or is `@clip`
-        if not os.path.isfile(fn) and fn != "@clip":
-            print(f"Error: input file \"{fn}\" does not exist")
-            sys.exit(1)
+    # check if input file exists
+    if not os.path.isfile(args.file) and args.file != "@clip":
+        print(f"Error: input file \"{args.file}\" does not exist")
+        sys.exit(1)
 
-    # read files
-    datas = []
-    for fn in args.file:
-        if fn == "@clip":
-            # clipboard
+    # @clip is clipboard input
+    if args.file == "@clip":
+        code = pyperclip.paste()
+    else:
+        with open(args.file, "r") as f:
+            code = f.read()
 
-            datas.append((fn, pyperclip.paste()))
+    try:
+        out = compile_code(code, args.file)
+    except Error as e:
+        e.print()
 
-        else:
-            # file
+        # print the traceback
+        if args.print_exceptions:
+            raise e
 
-            with open(fn, "r") as f:
-                datas.append((fn, f.read()))
-
-    # compile all files
-    outs = []
-    for data in datas:
-        # skip if already compiled
-        if data[0].endswith(".mind") or data[0].endswith(".masm"):
-            outs.append(data[1])
-            continue
-
-        try:
-            out = Preprocessor.preprocess(data[1])
-            out = Lexer(os.path.dirname(os.path.abspath(data[0]))).lex(out, data[0])
-            out = Parser().parse(out)
-            out = out.generate()
-            out = Optimizer.optimize(out)
-        except MlogError as e:
-            e.print()
-            if args.print_exceptions:
-                raise e
-            else:
-                sys.exit(1)
-
-        # add to compiled files
-        outs.append(out)
-
-    # link the compiled files
-    out = Linker.link(outs)
+        sys.exit(1)
 
     if output_method == IOMethod.FILE:
         # output to file
@@ -116,8 +100,9 @@ def main():
     elif output_method == IOMethod.STD:
         # output to stdout
 
-        # check if line numbers should be printed
         if vars(args)["lines"]:
+            # print line numbers
+
             lines = out.splitlines()
             max_line = len(str(len(lines)))
             for i, ln in enumerate(lines):
